C
C  ****************************************************************************
C  *  Reads 3D map files, radii range to use in the standard map, calls 3D    *
C  *  correlation routine.                                                    *
C  *                                                                          *
C  *  13-JAN-1992 BY HOLLAND CHENG                                            *
C  *  Use full map, not octant, D. Belnap, 23 May 1998                        *
C  *  MODIFIED  3-JUN-1998 BY TSB (PIF compatible).                           *
C  *  MODIFIED 2-AUG-2000 BY J. Folk  (WIN_SIZE VARIABLE)                     *
C  *  MODIFIED 4-DEC-2000 BY J. Folk  (ADDED QTR_SWITCH)                      *
C  ****************************************************************************
C
      SUBROUTINE CMP3D (ICHC, WIN_SIZE, QTR_SWITCH)
C
      IMPLICIT NONE
C
      include '../include/pifhead.inc'
      TYPE(PIFHEAD_T)  MAPHEAD, IMGHEAD, PRJHEAD, PFTHEAD, SFHEAD
      COMMON /PIFHEAD/ MAPHEAD, IMGHEAD, PRJHEAD, PFTHEAD, SFHEAD

      COMMON/MAPCEN/XC,YC,ZC
      COMMON/MAGRAD/RADLO,RADHI,RADSTEP,BMAG,IIX,IIY,IIZ
      INTEGER*4 NUM, I, J, K, NCOL1, NROW1, NSEC1, NCOL2, NROW2, NSEC2
      INTEGER*4 COL1, COL2, ROW1, ROW2, SEC1, SEC2, TITLE(18)
      INTEGER*4 COL21, COL22, ROW21, ROW22, SEC21, SEC22
      INTEGER*4 ICHC, IIX, IIY, IIZ, QTR_SWITCH
      INTEGER*4 NCOL, NROW, NSEC, MAXRAD, IRUN, WIN_SIZE
      REAL*4    TMP1, TMP2, TMP3, RADLO, RADHI, RADSTEP, STD
      REAL*4    XC2,YC2,ZC2,XC1,YC1,ZC1, XC, YC, ZC, BMAG
      REAL*4    MAP_MIN, MAP_MAX, MAP_AVG, MAP_STD
      character(len=1) READWRITE(4)
      CHARACTER MAP_FILENAME*64
      integer alloc_stat

      REAL*4, allocatable :: MAP1(:), MAP2(:), TMP_MAP(:)
C   ----------------------------------------------------------------
      PRINT *
      PRINT *, '*************************************************'
      PRINT *, '*      FIND BEST FIT BETWEEN 2 MAPS USING       *'
      PRINT *, '*  CORRELATION COEFFICIENTS AND LEAST SQUARES   *'
      PRINT *, '*-----------------------------------------------*'
      PRINT *, '*            ***** WARNING *****                *'
      PRINT *, '* Program assumes the centers of the 3D objects *'
      PRINT *, '* are in the center of the volume.              *'
      PRINT *, '*************************************************'
C ***********************************
C *  Set variables for subroutines  *
C ***********************************
      IIX = 1
      IIY = 1
      IIZ = 1
C  ************************
C  *  OPEN THE FIRST MAP  *
C  ************************
      PRINT *, '1st 3D MAP FILENAME: '
      READ (*,'(A)',ERR=999,END=999) MAP_FILENAME
      MAPHEAD%LUNUM = 1
      READWRITE(1)  = 'r'
      CALL PIF_OPEN (MAP_FILENAME, MAPHEAD%LUNUM, READWRITE) 
      CALL PIF_READ_GH (MAPHEAD)
      IF(MAPHEAD%MODE.EQ.20) THEN
         MAPHEAD%IMGNUM = 1                              ! LOAD UP "IMG" #1
         CALL PIF_READ_DH (MAPHEAD)
         NCOL1 = MAPHEAD%NUMCOLS
         NROW1 = MAPHEAD%NUMROWS
         NSEC1 = MAPHEAD%NUMSECS
       ELSE
         PRINT *, '***********************************'
         PRINT *, '* Invalid PIF mode for 1st 3D MAP *'
         CALL TSBEND
      ENDIF
      IF(NSEC1.LE.1) THEN
         allocate(map1(ncol1*nrow1), stat=alloc_stat)
         if(alloc_stat > 0) then
            write(*,*) 'Unable to allocate MAP1 in CMP3D'
            go to 999
         endif
         CALL PIF_READ_WMAPI2 (MAPHEAD, MAP1, NCOL1, NROW1, 1)
         CALL PIF_CLOSE (1)
       ELSE
         COL1     = 1
         ROW1     = 1
         SEC1     = 1
C  *********************************
C  *  Get center for standard map  *
C  *********************************
         XC1  = REAL(NCOL1+1)/2.0
         YC1  = REAL(NROW1+1)/2.0
         ZC1  = REAL(NSEC1+1)/2.0
         XC   = XC1
         YC   = YC1
         ZC   = ZC1
         TMP1 = 0.0
         TMP2 = 0.0
         TMP3 = 0.0
 400     WRITE (6,'(A,2(F8.2,A),F8.2,A)') ' MAP CENTER (<CR>=', XC, ',',
     1   YC, ',', ZC, '):'
         READ(*,*,ERR=400) TMP1, TMP2, TMP3
         IF(TMP1.GT.0.) XC = TMP1
         IF(TMP2.GT.0.) YC = TMP2
         IF(TMP3.GT.0.) ZC = TMP3
C  *******************************************
C  *  Load the standard map into MAP1 array  *
C  *******************************************
         allocate(map1(ncol1*nrow1*nsec1), stat=alloc_stat)
         if(alloc_stat > 0) then
            write(*,*) 'Unable to allocate MAP1 in CMP3D'
            go to 999
         endif
C          ****************************
C----------*  Load 1st MAP into MAP1  *
C          ****************************
         CALL PIF_READ_WMAPI2 (MAPHEAD, MAP1, NCOL1, NROW1, NSEC1)
         CALL MAP_STATS (1, MAP1, NCOL1, NROW1, NSEC1, MAP_MIN, MAP_MAX, 
     1   MAP_AVG, MAP_STD, 0)
      ENDIF
      STD = MAP_STD
      CALL PIF_CLOSE (1)
C  *************************
C  *  OPEN THE SECOND MAP  *
C  *************************
      PRINT *, '2nd 3D MAP FILENAME: '
      READ (*,'(A)',ERR=999,END=999) MAP_FILENAME
      MAPHEAD%LUNUM = 1
      READWRITE(1)  = 'r'
      CALL PIF_OPEN (MAP_FILENAME, MAPHEAD%LUNUM, READWRITE) 
      CALL PIF_READ_GH (MAPHEAD)
      IF(MAPHEAD%MODE.EQ.20) THEN
         MAPHEAD%IMGNUM = 1                              ! LOAD UP "IMG" #1
         CALL PIF_READ_DH (MAPHEAD)
         NCOL2 = MAPHEAD%NUMCOLS
         NROW2 = MAPHEAD%NUMROWS
         NSEC2 = MAPHEAD%NUMSECS
       ELSE
         PRINT *, '***********************************'
         PRINT *, '* Invalid PIF mode for 2nd 3D MAP *'
         CALL TSBEND
      ENDIF
      IF(NSEC2.LE.1) THEN
         allocate(map2(ncol2*nrow2), stat=alloc_stat)
         if(alloc_stat > 0) then
            write(*,*) 'Unable to allocate MAP2 in CMP3D'
            go to 999
         endif
         CALL PIF_READ_WMAPI2 (MAPHEAD, MAP2, NCOL2, NROW2, 1)
         CALL PIF_CLOSE (1)
       ELSE
         allocate(map2(ncol2*nrow2*nsec2), stat=alloc_stat)
         if(alloc_stat > 0) then
            write(*,*) 'Unable to allocate MAP2 in CMP3D'
            go to 999
         endif
C          ****************************
C----------*  Load 2nd MAP into MAP2  *
C          ****************************
         CALL PIF_READ_WMAPI2 (MAPHEAD, MAP2, NCOL2, NROW2,
     1   NSEC2)
         CALL MAP_STATS (1, MAP2, NCOL2, NROW2, NSEC2, 
     1   MAP_MIN, MAP_MAX, MAP_AVG, MAP_STD, 0)

C  *********************************************************************
C  * If MAP1 larger than MAP2, add extra null pixels to periphery of   *
C  * MAP2 to give it the same dimensions as MAP1 without moving center * 
C  ********************************************************************* 

      IF (NCOL2.LT.NCOL1.OR.NROW2.LT.NROW1.OR.NSEC2.LT.NSEC1) THEN 
         PRINT *, 'MAP2 has smaller dimensions than MAP1.'
         PRINT *, 'Applying padding routine...'  
         allocate(tmp_map(ncol1*nrow1*nsec1), stat=alloc_stat)
         if(alloc_stat > 0) then
            write(*,*) 'Unable to allocate TMP_MAP in CMP3D'
            go to 999
         endif
         CALL ADD_MAP_PADDED(MAP2,NCOL2,NROW2,NSEC2,TMP_MAP,
     1  NCOL1, NROW1, NSEC1)
         NCOL2 = NCOL1
         NROW2 = NROW1
         NSEC2 = NSEC1 
         deallocate(map2)
         allocate(map2(ncol1*nrow1*nsec1), stat=alloc_stat)
         if(alloc_stat > 0) then
            write(*,*) 'Unable to allocate MAP2 in CMP3D'
            go to 999
         endif
         MAP2 = TMP_MAP ! Fortran 90 array syntax
      ENDIF

C  *******************************
C  *  Get center for second map  *
C  *******************************
         XC2  = REAL(NCOL2+1)/2.0
         YC2  = REAL(NROW2+1)/2.0
         ZC2  = REAL(NSEC2+1)/2.0
         TMP1 = 0.0
         TMP2 = 0.0
         TMP3 = 0.0
 500     WRITE (6,'(A,2(F8.2,A),F8.2,A)') ' MAP CENTER (<CR>=', XC2, 
     1        ',', YC2, ',', ZC2, '):'
         READ(*,*,ERR=500) TMP1, TMP2, TMP3
         IF(TMP1.GT.0.) XC2 = TMP1
         IF(TMP2.GT.0.) YC2 = TMP2
         IF(TMP3.GT.0.) ZC2 = TMP3
      ENDIF
      CALL PIF_CLOSE (1)
      STD = (STD + MAP_STD)/2.0
C  **************************************
C  *  Get Radii range for standard map  *
C  **************************************
         RADLO   = 0.0
         RADHI   = REAL(NCOL1-1)/2.0
         RADSTEP = 1.0
         TMP1    = 0.0
         TMP2    = 0.0
         TMP3    = 0.0
  800         PRINT *, ' For Standard Map,',
     1         'Enter inner and outer radius, and step (<CR>= ',
     2         RADLO, RADHI, RADSTEP, '):'
         READ (*,*,ERR=800) TMP1, TMP2, TMP3
         IF(TMP1.GT.0.0) RADLO   = TMP1
         IF(TMP2.GT.0.0) RADHI   = TMP2
         IF(TMP3.GT.0.0) RADSTEP = TMP3
         IF(RADLO.GT.RADHI) THEN
           PRINT *, '*****************************'
           PRINT *, '* RADMIN cannot be > RADMAX *'
           PRINT *, '*****************************'
           GO TO 800
         ENDIF
C       ******************************
C       *  Output radii information  *
C       ******************************
      PRINT *, ' RADLO, RADHI, RADSTEP =',
     1 RADLO, RADHI, RADSTEP
C  ******************************
C  *  Call correlation routine  *
C  ******************************
 900      CALL CC_3D (MAP1, MAP2, NCOL1, NROW1, NSEC1, NCOL2, NROW2, 
     1 NSEC2, ICHC, STD, WIN_SIZE, QTR_SWITCH)
C  *******************************************************************
C  *  Allow user to repeat correlations with a different mag. range  *
C  *******************************************************************
      PRINT *, ' Change magnification range? (<CR>=No/<1>=Yes): '
      READ (*,*) IRUN
      IF(IRUN.EQ.1) GO TO 900
C  ***********************************
C  *  Clear memory and exit routine  *
C  ***********************************
      deallocate(map1, map2)
      if(allocated(tmp_map)) deallocate(tmp_map)
 999  RETURN
      END
C
C  ****************************************************************************
C  *  Computes correlation coefficients between two 3D maps at range of mag.  *
C  *  factors.                                                                *
C  *                                                                          *
C  *  14-JAN-1992 BY HOLLAND CHENG, AS SUBROUTINE "CCOMAP3DS"                 *
C  *  18-OCT-1997 fixed array overflow, NTOT, index error, T.Baker, D.Belnap  *
C  *  23-MAY-1998 added comments, minor changes, D. Belnap                    *
C  *  3-JUN-1998 BY TSB (PIF compatible).                                     *
C  *  12-MAY-2000 Fixed RADHISQ = (RADHI+RADSTEP)**2 T. Baker, J. Folk        * 
C  ****************************************************************************
C
      SUBROUTINE CC_3D (MAP, MAP2, NCOL1, NROW1, NSEC1, NCOL2, NROW2, 
     1 NSEC2, ICHC, STD,
     1  xc,yc,zc,
     1  mag_cen,mag_step,mag_num,
     2  radlo,radhi,radstep,bmag,iix,iiy,iiz, 
     2  mags,coefs,subMap2d3d, qtr_switch)
C
      IMPLICIT NONE
C
Ca      COMMON/MAPCEN/XC,YC,ZC
Ca      COMMON/MAGRAD/RADLO,RADHI,RADSTEP,BMAG,IIX,IIY,IIZ
C
      INTEGER*4 NCOL1, NROW1, NSEC1, NCOL2, NROW2, NSEC2
      INTEGER*4 I, J, K, L, NUM, NCOL, NROW, NSEC, OFFSET, IB, ICHC
      INTEGER*4 MAG_NUM, NMAG, II, IIX, IIY, IIZ, LIM, NBINS, BIN, LPOL
Ca      PARAMETER (LIM=501,LPOL=6)
      PARAMETER (LIM=1001,LPOL=6)
      INTEGER*4 NTOT(LIM), NFIT, TNTOT, win_size, QTR_SWITCH
      INTEGER*4 ITEST, twod_switch, array_tot, subMap2d3d, nsec_local
      REAL*4    TMP1, TMP2, TMP3, RADIB(lim), BMAG
      REAL*4    MAG, MAG_CEN, MAG_STEP, TCCOEF, TBCCOEF
      REAL*4    AVE,ADEV,SDEV,VAR,SKEW,CURT,MAGCCOEF(LIM), STD
      REAL*4    RADLO,RADHI,RADDSQ,XC,YC,ZC,RADLOSQ,RADHISQ,YSQ,ZSQ
      REAL*4    MAP(NCOL1,NROW1,NSEC1), A, B, RADSTEP, AA, BB
      REAL*4    CCOEF(LIM), CCOEF2(LIM), TCCOEF2, TBCCOEF2
      real*4, allocatable :: rdens(:), wrdens(:), wccoef(:), s_wccoef(:)
      real*4, allocatable :: CNTR1(:), CNTR2(:)
      real*4    mags(mag_num),coefs(mag_num)
      REAL*4    BMAG2, MAGCCOEF2(LIM), TOT, MAP2(NCOL2,NROW2,NSEC2)
      REAL*8    TSUMA, TSUMB, TSUMASQ, TSUMBSQ, TSUMAB, SUMAB(LIM)
      REAL*8    SUMA(LIM), SUMB(LIM), SUMASQ(LIM), SUMBSQ(LIM)
      integer   alloc_stat

C  -------------------------------------------------------------------
C 
C ************************************************
C *  Get number of bins between RADHI and RADLO  *
C ************************************************
      NBINS = 1 + NINT((RADHI-RADLO)/RADSTEP)
      IF(NBINS.LE.0) NBINS = 1
      IF(NBINS.GT.LIM) THEN
        PRINT *, '*****************************'
        PRINT *, '* RADSTEP must be increased *'
        PRINT *, '*****************************'
        GO TO 999
      ENDIF
      RADLOSQ = RADLO*RADLO
      RADHISQ = (RADHI+RADSTEP)**2
C **********************************************************
C *  Set NCOL, NROW, NSEC to maximum NCOL of the two maps  *
C **********************************************************
      array_tot = 1   ! Array counter
      NCOL = MAX (NCOL1, NCOL2)
      NROW = NCOL
      NSEC = NCOL
C**********************************************
C* Reset NSEC, NROW if using only quarter map *
C**********************************************
       IF (QTR_SWITCH.EQ.1) THEN
        ITEST = INT(NCOL1/2) + 1
        NSEC = INT(NSEC1/2) + 1
        NROW = INT(NROW1/2) + 1
       ENDIF
C ********************
C *  Get mag. range  *
C ********************
Ca        PRINT *, 'MAG_CEN,MAG_STEP,MAG_NUM (2F,I)',
Ca      1 ' <CR>=1.0,0.01,11: '
Ca        READ (*,'(2F,I)',ERR=999,END=999) MAG_CEN, MAG_STEP, MAG_NUM
Ca      IF(MAG_NUM.LE.0) MAG_NUM = 11
Ca        IF(MOD(MAG_NUM,2).EQ.0) MAG_NUM = MAG_NUM + 1
Ca        IF(MAG_CEN.LE.0.0)   MAG_CEN = 1.0
Ca      IF(MAG_STEP.LE.0.0) MAG_STEP = 0.01
C **************************
C *  Initialize variables  *
C **************************
      MAG     = MAG_CEN - (((MAG_NUM-1)/2)*MAG_STEP)
      NMAG    = 0
      TCCOEF  = 0.0
      TBCCOEF = 0.0
      TBCCOEF2= 0.0
      BMAG    = 1.0
      BMAG2   = 1.0

Ca ***********************************************
Ca * Setup for 2D processing
Ca ***********************************************
      nsec_local = nsec
Ca      Check to see if we're doing 2d or 3d algorithm
      if(subMap2d3d .eq. 2) then
         iiz = zc
         nsec_local = zc+1
      endif

C ******************************************
C *  Determine Cor. Coef. over mag. range  *
C ******************************************
      DO II = 1,MAG_NUM
C         ****************************************
C         *  Initialize variables for each mag.  *
C         ****************************************
        TCCOEF2 = 0.0
        TSUMA   = 0.0
        TSUMB   = 0.0
        TSUMAB  = 0.0
        TSUMASQ = 0.0
        TSUMBSQ = 0.0
        NFIT    = 0
        TNTOT   = 0
C         ************************************
C         *  Set sums to zero for each mag.  *
C         ************************************
        DO IB = 1,NBINS
          SUMA(IB)    = 0.0
          SUMB(IB)    = 0.0
          SUMAB(IB)   = 0.0
          SUMASQ(IB)  = 0.0
          SUMBSQ(IB)  = 0.0
          NTOT(IB)    = 0
        ENDDO
C       ******************************************************
C       *  If voxel in selected radii, interpolate, and sum  *
C       ******************************************************
C
        DO K = iiz,NSEC_local
          ZSQ = (K-ZC)**2
          DO J = iiy,NROW
            YSQ = (J-YC)**2
            DO I = iix,NCOL
             IF (K.EQ.NSEC_local.OR.J.EQ.NROW) THEN
            IF (QTR_SWITCH.EQ.1) THEN
              IF (I.GT.ITEST) GOTO 53
            ENDIF
             ENDIF
              RADDSQ = (I-XC)**2 + YSQ + ZSQ
              IF(RADDSQ.GE.RADLOSQ .AND. RADDSQ.LE.RADHISQ) THEN
                B = MAP(I,J,K)
                CALL MAG3D_INTRP (A, MAP2, NCOL1, NROW1, NSEC1, I, J, 
     1          K, MAG, XC, YC, ZC)
              BIN = INT((SQRT(RADDSQ) - RADLO) / RADSTEP) + 1
              IF(BIN.GT.NBINS) GOTO 53 
              SUMA(BIN)   = SUMA(BIN)   + A
              SUMB(BIN)   = SUMB(BIN)   + B
              SUMAB(BIN)  = SUMAB(BIN)  + A*B
              SUMASQ(BIN) = SUMASQ(BIN) + A*A
              SUMBSQ(BIN) = SUMBSQ(BIN) + B*B
                NTOT(BIN)   = NTOT(BIN)   + 1
                NFIT        = NFIT        + 1
               IF(BIN.EQ.1) THEN
               ENDIF
              ENDIF
53            ENDDO
          ENDDO
        ENDDO
C
C       *******************************
C       *  Compute Correlation Coef.  *
C       *******************************
      DO 303 IB = 1,NBINS
        CCOEF2(IB) = 0.
        IF (NTOT(IB).EQ.0) THEN                ! TRAPPING OVERFLOWING 
          CCOEF(IB) = 0.
          GO TO 303
        ENDIF
        TOT = REAL(NTOT(IB))
        IF ((SUMASQ(IB) - (SUMA(IB)*SUMA(IB))/TOT).LE.0. .OR.
     1     (SUMBSQ(IB) - (SUMB(IB)*SUMB(IB))/TOT).LE.0.) THEN 
           CCOEF(IB) = 0.
         ELSE
           CCOEF(IB) = (   SUMAB(IB) - (SUMA(IB)*SUMB(IB))/TOT) /
     1             (SQRT(SUMASQ(IB) - (SUMA(IB)*SUMA(IB))/TOT)*
     2              SQRT(SUMBSQ(IB) - (SUMB(IB)*SUMB(IB))/TOT))
           CCOEF2(IB)= CCOEF(IB)*CCOEF(IB) 
           TCCOEF2   = TCCOEF2 + CCOEF2(IB) 
           RADIB(IB) = RADLO   + (IB-1)*RADSTEP
           TSUMA     = TSUMA   + SUMA(IB)    
           TSUMB     = TSUMB   + SUMB(IB)
           TSUMAB    = TSUMAB  + SUMAB(IB)   
           TSUMASQ   = TSUMASQ + SUMASQ(IB) 
           TSUMBSQ   = TSUMBSQ + SUMBSQ(IB)  
           TNTOT     = TNTOT   + NTOT(IB)  
           IF(MAG_NUM.NE.1 .AND. IB.EQ.1) WRITE(9,'(2F10.4,A)') 
     1     RADIB(IB), CCOEF(IB), '  MOV'
       ENDIF
 303      CONTINUE
      IF(TNTOT.LE.0) THEN                 ! TRAP FOR OVERFLOWING
        PRINT *
        PRINT *, 'TNTOT=0,  @MAG: ',MAG
        GO TO 900
      ENDIF
      IF ((TSUMASQ - (TSUMA*TSUMA)/TNTOT).LE.0. .OR.
     2   (TSUMBSQ - (TSUMB*TSUMB)/TNTOT).LE.0.) THEN 
        PRINT *
        PRINT *, 'OVERFLOW IN CCOEF CALC,  @MAG: ',MAG
        GO TO 900
      ENDIF
C
        TCCOEF = (TSUMAB - (TSUMA*TSUMB)/REAL(TNTOT)) /
     1           (SQRT(TSUMASQ - (TSUMA*TSUMA)/REAL(TNTOT))*
     2            SQRT(TSUMBSQ - (TSUMB*TSUMB)/REAL(TNTOT)))
        NMAG             = NMAG + 1
        MAGCCOEF (NMAG)  = TCCOEF 
        MAGCCOEF2 (NMAG) = TCCOEF2 
        IF (ABS(TCCOEF).GT.ABS(TBCCOEF)) THEN
          TBCCOEF = TCCOEF
          BMAG    = MAG
        ENDIF
        IF (TCCOEF2.GT.TBCCOEF2) THEN
          TBCCOEF2 = TCCOEF2
          BMAG2    = MAG
        ENDIF
Ca        PRINT *, ' CCOEF, MAG, sq =', TCCOEF,MAG,TCCOEF2
        mags(array_tot)    = mag
        coefs(array_tot)   = tccoef
        array_tot = array_tot + 1
  900        MAG = MAG + MAG_STEP
      ENDDO
C ********************************************************
C *  Output best mag. (if a range) and other parameters  *
C ********************************************************
      IF(NMAG.GT.2) THEN
Ca        PRINT *,' BEST CCOEF, MAG =', TBCCOEF, BMAG
Ca        PRINT *, ' BEST CCsq, MAG2 =', TBCCOEF2, BMAG2
        CALL MOMENT295 (MAGCCOEF,NMAG,AVE,ADEV,SDEV,VAR,SKEW,CURT)
Ca        PRINT *, ' AVG, SD, VAR: ', AVE, SDEV, VAR
Ca        PRINT *
      ENDIF
C ****************************************************************
C *  If user chose to also scale densities, then call CCOMAP3CS  *
C ****************************************************************
      IF(ICHC.EQ.3) THEN
         allocate(cntr1(nfit), cntr2(nfit), rdens(nbins), wrdens(nbins),
     $        wccoef(nbins), s_wccoef(nbins), stat=alloc_stat)
        if(alloc_stat > 0) then
           write(*,*) ' ERR>  NOT ENOUGH MEMORY FOR CONTRAST SCALING !'
           write(*,*)
           write(*,*)
           go to 999
        endif

        IF (win_size.LE.0) win_size = 4
        twod_switch = 3         ! 3 dimensional
        CALL CCOMAP3CS (MAP, MAP2, NCOL1, NROW1, NSEC1, TBCCOEF, 
     1       NCOL2, NROW2, NSEC2, NFIT, CNTR1, CNTR2, LPOL, STD,
     2       XC,YC,ZC,RADLO,RADHI,RADSTEP,BMAG,rdens,wrdens,
     3       ccoef,wccoef,s_wccoef,nbins,BB,AA, win_size, QTR_SWITCH,
     4       twod_switch)
        deallocate(cntr1, cntr2, rdens, wrdens, wccoef, s_wccoef)

      ENDIF
      RETURN
C ***************
C *    ERROR    *
C ***************
  999      PRINT *, '** BUG ALERT!  pls send a log to [rhc] **'
      CALL TSBEND
      END
C
C****************************************************************************
C*  SUBROUTINE CCOMAP3CS 
C*  Modified 5/25/00 RWA - Added args necessary to avoid using common
C*  block to pass.
C*  Modified 7/27/00 J. Folk - Improved algorithm for determining recommended
C*  RADLO and RADHI values, using S_WCCOEF to find the largest "bunch" of
C*  continuous nonzero values, then within that range of radii finding the
C*  peak value of WCCOEF, and searching radii less than that peak value
C*  to find RADLO, and searching radii greater than that peak value to find
C*  RADHI.
C****************************************************************************
      SUBROUTINE CCOMAP3CS (MAP, MAP2, NCOL1, NROW1, NSEC1, TBCCOEF, 
     1 NCOL2, NROW2, NSEC2, NFIT, CNTR1, CNTR2, LPOL, MAP_STD,
     2 XC,YC,ZC,RADLO,RADHI,RADSTEP,BMAG,
     3 rdens,wrdens,ccoef,wccoef,s_wccoef,nbins, BB, AA, win_size,
     4 QTR_SWITCH, twod_switch, threshValue, useAbs)
C
C      14-JAN-1992 BY HOLLAND CHENG 
c      last revision 31-JAN-95 [RHC]
C      18-OCT-1997 fixed array overflow, index error, etc., T.Baker, D.Belnap
C       11-MAY-2000 print out rad and cc to screen as well as file, Jon Folk
C       12-MAY-2000 commented out maps of different sizes, added section to
C                    print out rad and ccoef for all radii, Jon Folk          
C       12-MAY-2000 corrected bugs, fixed RADHISQ = (RADHI+RADSTEP)**2,
C                    fixed error in highest radius ccoef(ib), T. Baker, J. Folk 
C       22-MAY-2000 added weighted average using cmp_map_1d to calculate
C                   circular average normalized to 1.0 and setting all
C                   negative density to 0.0.  J. Folk
      IMPLICIT NONE
C
Ca      COMMON/MAPCEN/XC,YC,ZC
Ca      COMMON/MAGRAD/RADLO,RADHI,RADSTEP,BMAG,IIX,IIY,IIZ
C
      INTEGER*4 NCOL1, NROW1, NSEC1, NCOL2, NROW2, NSEC2
      INTEGER*4 I, J, K, L, NUM, NCOL, NROW, NSEC, OFFSET
      INTEGER*4 NMAG, II, IIX, IIY, IIZ, LIM, NBINS, BIN, IIC
      INTEGER*4 LPOL, LOWTAG, RADTAG, NUM_DIGIT_R4, IB, NFIT, MWT
      INTEGER*4 D0, D1, D2, D3, D4, D5, D6
      INTEGER*4 NRDENS, WIN_SIZE
      INTEGER*4 BUNCH, MAX_BUNCH, QTR_SWITCH
      INTEGER*4 ITEST, twod_switch, useAbs, isOK
      PARAMETER (LIM=501)
      integer alloc_stat
      REAL*4    MAX_BUNCH_RADLO, MAX_BUNCH_RADHI
      REAL*4    MAX_BUNCH_PEAK, MAX_BUNCH_PEAKIB
Ca      REAL*4    WCCOEF(LIM), RDENS(1), WRDENS(1)
      REAL*4    threshValue, absThreshValue
      REAL*4    WCCOEF(nbins), RDENS(nbins), WRDENS(nbins)
      REAL*4    CCOEF(nbins), max_s_wccoef
      REAL*4    RSQ_LIM, WCCOEF_MAX, S_WCCOEF(nbins)      
      REAL*4    CNTR1(NFIT), CNTR2(NFIT), MAG, MAP_STD
      REAL*4    AA, BB, UNA, UNB, CHISQ, GDNS, TCCOEF, TBCCOEF, BMAG
      REAL*4    AVE,ADEV,SDEV,VAR,SKEW,CURT,MAGCCOEF(LIM)
      REAL*4    RADLO,RADHI,RADDSQ,XC,YC,ZC,RADLOSQ,RADHISQ,YSQ,ZSQ
      REAL*4    MAP(NCOL1,NROW1,NSEC1),A,B,RADSTEP
      REAL*4    MAP2(NCOL2,NROW2,NSEC2), S, P
      REAL*4    TNTOT, NTOT(LIM), RADIB, TMPLO, TMPHI, TMP1, TMP2
      REAL*8    TSUMA, TSUMB, TSUMASQ, TSUMBSQ, TSUMAB, SUMAB(LIM)
      REAL*8    SUMA(LIM), SUMB(LIM), SUMASQ(LIM), SUMBSQ(LIM)
      integer*4, allocatable :: rdens_num(:)
      real*4, allocatable    :: rdens_std(:)

C  -------------------------------------------------------------------
C
Ca      RWA - for now let's just init these - we'll pas them in later???
      absThreshValue = abs(threshValue)
      iix = 1
      iiy = 1
      iiz = 1
      MAG     = BMAG 
      LOWTAG  = 0 
      RADTAG  = 0 
      TMPLO   = RADLO
      TMPHI   = RADHI
      NCOL    = MAX (NCOL1, NCOL2)
      NROW    = NCOL
Ca  Careful!!! What about twod_switch???      NSEC    = NCOL
      NSEC = NCOL
      if(twod_switch .eq. 2) then
         nsec = 1
      endif
      WCCOEF_MAX = 0.0

C  *************************************************
C  * Reset NSEC and NROW if only using quarter map *
C  *************************************************
       IF (QTR_SWITCH.EQ.1) THEN 
        ITEST = INT(NCOL1/2) + 1
        NSEC = INT(NSEC1/2) + 1
        NROW = INT(NROW1/2) + 1
       ENDIF

C  ************************************************************
C  *    PRINT TO SCREEN AND FILE RAD, CCOEF,NTOT FOR ALL RADII* 
C  *    NOTE: ASSUMES NCOL1=NCOL2 AND NROW1=NROW2 
C  ************************************************************
       RADHISQ = (REAL((NCOL-1)/2)+RADSTEP)**2
      RSQ_LIM = RADHISQ  ! USED FOR CMP_MAP_1D 
Ca Pass these in       NBINS = 1 + NINT(((NCOL-1)/2)/RADSTEP)
      NRDENS = NBINS ! USED FOR CMP_MAP_1D
C  ************************************************************
C  *  CALL CMP_MAP_1D TO USE CIRCULAR DENSITY AVERAGING TO    *
C  *  AND NORMALIZATION TO 1.0 AND REMOVAL OF NEGATIVE        *
C  *  DENSITIES CALCULATE WRDENS, WHICH IS LATER USED TO      *
C  *  CALCULATE WCCOEF, A WEIGHTED CORRELATION COEFFICIENT.   *
C  ************************************************************
      allocate(rdens_num(nrdens), rdens_std(nrdens),
     $     stat = alloc_stat)
      if(alloc_stat > 0) then
         write(*,*) 'Unable to allocate memory in CCOMAP3CS'
      endif

      CALL CMP_MAP_1D (MAP, NCOL1, NROW1, NSEC1, XC, YC, ZC, 
     1        NRDENS, RSQ_LIM, RDENS, WRDENS, RDENS_NUM, RDENS_STD,
     2  QTR_SWITCH)

C  ****************************************
C  *  CALCULATE CORRELATION COEFFICIENT   *
C  ****************************************
        DO IB = 1,NBINS
          SUMA(IB)    = 0.0
          SUMB(IB)    = 0.0
          SUMAB(IB)   = 0.0
          SUMASQ(IB)  = 0.0
          SUMBSQ(IB)  = 0.0
          NTOT(IB)    = 0.0
        ENDDO
        DO K = iiz,NSEC
          ZSQ = (K-ZC)**2
          DO J = iiy,NROW
            YSQ = (J-YC)**2
            DO 9 I = iix,NCOL
           IF (K.EQ.NSEC.OR.J.EQ.NROW) THEN
            IF(QTR_SWITCH.EQ.1) THEN
            IF(I.GT.ITEST) GOTO 9 
            ENDIF
           ENDIF
              RADDSQ = (I-XC)**2 + YSQ + ZSQ
              IF(RADDSQ.GT.RADHISQ) GOTO 9
                B = MAP(I,J,K)
                CALL MAG3D_INTRP (A, MAP2, NCOL1, NROW1, NSEC1, I, J,
     1 K, MAG, XC, YC, ZC)
            BIN = INT(SQRT(RADDSQ) / RADSTEP) + 1
            IF(BIN.GT.NBINS) GOTO 9 
            SUMA(BIN)   = SUMA(BIN)   + A
            SUMB(BIN)   = SUMB(BIN)   + B
            SUMAB(BIN)  = SUMAB(BIN)  + A*B
            SUMASQ(BIN) = SUMASQ(BIN) + A*A
            SUMBSQ(BIN) = SUMBSQ(BIN) + B*B
            NTOT(BIN)   = NTOT(BIN)   + 1.0
9           ENDDO
          ENDDO
        ENDDO

Ca        PRINT *, 'RADIUS, CCOEF, WCCOEF FOR ALL RADII'
        DO 10 IB = 1,NBINS
         CCOEF(IB) = 0.
         RADIB = (IB-1)*RADSTEP
         IF (NTOT(IB).EQ.0.) THEN
           WCCOEF(IB) = CCOEF(IB)*WRDENS(IB)
         IF (WCCOEF(IB).LT.0.0) WCCOEF(IB) = 0.0
           GOTO 10
         ENDIF
         IF ((SUMASQ(IB) - (SUMA(IB)*SUMA(IB))/NTOT(IB)).LE.0. .OR.
     1 (SUMBSQ(IB) - (SUMB(IB)*SUMB(IB))/NTOT(IB)).LE.0.) THEN
           WCCOEF(IB) = CCOEF(IB)*WRDENS(IB)
          IF (WCCOEF(IB).LT.0.0) WCCOEF(IB) = 0.0
         ELSE
           CCOEF(IB) = (SUMAB(IB) - (SUMA(IB)*SUMB(IB))/NTOT(IB)) /
     1 (SQRT( SUMASQ(IB) - (SUMA(IB)*SUMA(IB))/NTOT(IB))*
     2 SQRT(SUMBSQ(IB) - (SUMB(IB)*SUMB(IB))/NTOT(IB)))
           WCCOEF(IB) = CCOEF(IB)*WRDENS(IB)
          IF (WCCOEF(IB).LT.0.0) WCCOEF(IB) = 0.0
         ENDIF
      WCCOEF_MAX = MAX(WCCOEF_MAX, WCCOEF(IB)) 
10      ENDDO 
      DO IB = 1,NBINS
        IF(WCCOEF(IB).LT.(0.01*WCCOEF_MAX)) WCCOEF(IB) = 0.0
          RADIB = (IB-1)*RADSTEP
        IF(IB.GT.2) THEN
          IF((WCCOEF(IB-2).EQ.0.0).AND.(WCCOEF(IB).EQ.0.0)) THEN
            WCCOEF(IB-1) = 0.0
          ENDIF
        ENDIF
        IF(IB.GT.1) THEN
Ca            PRINT *, (RADIB-1),CCOEF(IB-1),WCCOEF(IB-1)
        ENDIF
        IF(IB.EQ.NBINS) THEN
Ca            PRINT *,  RADIB,CCOEF(IB),WCCOEF(IB) 
        ENDIF
      ENDDO      
C ********************************
C *   SLIDING WINDOW ALGORITHM   *
C ********************************

      DO IB = 1, NRDENS
        S_WCCOEF(IB) = 0.0
      ENDDO
Ca
Ca
Ca      max_s_wccoef = 0.0
Ca      DO IB = NRDENS, 1, -1 
Ca        IF (IB.GE.WIN_SIZE) THEN
Ca           DO J = 1, WIN_SIZE
Ca           S_WCCOEF(IB) = S_WCCOEF(IB) + WCCOEF(IB-(J-1))
Ca         ENDDO
Ca        ENDIF
Ca        S_WCCOEF(IB) = S_WCCOEF(IB) / WIN_SIZE
Ca        max_s_wccoef = MAX(s_wccoef(IB),max_s_wccoef)
Ca      ENDDO       
Ca      DO IB = 1, NRDENS
Ca        RADIB = (IB-1) * RADSTEP
Ca      ENDDO 
C  ********************************************
C  * Use the S_WCCOEF, which is smoothed by   *
C  * the sliding window, to determine the     *
C  * largest consecutive series of nonzero    *
C  * S_WCCOEF values.  Use this to determine  *
C  * recommended RADLO and RADHI values.      *
C  ********************************************
Ca      BUNCH = 0
Ca      MAX_BUNCH = 0
Ca      MAX_BUNCH_RADLO  = 0.0
Ca      DO IB = NBINS, 1, -1
Ca        RADIB = (IB-1) * RADSTEP
Ca        IF(S_WCCOEF(IB).GT.(max_s_wccoef*.01)) THEN
Ca          BUNCH = BUNCH + 1
Ca          IF(BUNCH.GT.MAX_BUNCH) THEN
Ca            MAX_BUNCH = BUNCH
Ca            MAX_BUNCH_RADLO = RADIB
Ca            MAX_BUNCH_RADHI = MAX_BUNCH_RADLO + (MAX_BUNCH -
Ca     1 (WIN_SIZE - 1)) 
Ca          ENDIF
Ca        ELSE
Ca          BUNCH = 0
Ca        ENDIF
Ca      ENDDO

C**************************************************
C*  Now, with the info on the bunch location,
C*  search within that range of radii for the peak
C*  WCCOEF value.
C**************************************************


Ca      MAX_BUNCH_PEAK = 0.0
Ca      DO IB = 1, NBINS
Ca        RADIB = (IB - 1) * RADSTEP
Ca        IF ((RADIB.GE.MAX_BUNCH_RADLO) .AND. 
Ca     1 (RADIB.LE.MAX_BUNCH_RADHI)) THEN
Ca          IF(WCCOEF(IB).GT.MAX_BUNCH_PEAK) THEN
Ca            MAX_BUNCH_PEAK = WCCOEF(IB)
Ca            MAX_BUNCH_PEAKIB = IB
Ca          ENDIF
Ca        ENDIF
Ca      ENDDO

C**********************************************************
C*  Next, step to lower radii from the peak until the last
C*  nonzero WCCOEF value, this is the new
C*  recommended RADLO.  Then step to higher
C*  radii from the peak to the last nonzero
C*  WCCOEF value, this is the new recommended
C*  RADHI.
C***********************************************************
 
Ca      DO IB = MAX_BUNCH_PEAKIB, 1, -1
Ca        RADIB = (IB-1) * RADSTEP
Ca        IF (WCCOEF(IB).LE.(max_s_wccoef*.01)) GOTO 68
Ca        RADLO =  RADIB
Ca      ENDDO

68      IB = 0
Ca68      DO IB = MAX_BUNCH_PEAKIB, NBINS
Ca        RADIB = (IB-1) * RADSTEP
Ca        IF (WCCOEF(IB).LE.(max_s_wccoef*.01)) GOTO 66
Ca        RADHI = RADIB
Ca      ENDDO      


66      tmp1 = 0.
Ca66      PRINT *, 'RECOMMENDED NEW RADLO,RADHI', RADLO, RADHI
Ca      PRINT *, 'DESIRED RADLO, RADHI<',RADLO, RADHI, '>:'
Ca      tmp1 = 0.
Ca      tmp2 = 0.
Ca      READ (*,'(2F)') TMP1, TMP2
Ca      IF (TMP1.GT.0.) RADLO = TMP1
Ca      IF (TMP2.GT.0.) RADHI = TMP2 

      
C
Ca We are only going once now
Ca 1111      NBINS = 1 + NINT((RADHI-RADLO)/RADSTEP)
      NBINS = 1 + NINT((RADHI-RADLO)/RADSTEP)
      IF(NBINS.LE.0) NBINS = 1
      RADLOSQ = RADLO*RADLO
      RADHISQ = (RADHI+RADSTEP)**2 
      DO IB = 1,NBINS
        SUMA(IB)    = 0.0
        SUMB(IB)    = 0.0
        SUMAB(IB)   = 0.0
        SUMASQ(IB)  = 0.0
        SUMBSQ(IB)  = 0.0
        NTOT(IB)    = 0.0
      ENDDO
C
      IIC = 0 
C
      DO K = iiz,NSEC
        ZSQ = (K-ZC)**2
        DO J = iiy,NROW
          YSQ = (J-YC)**2
          DO I = iix,NCOL
            IF (QTR_SWITCH.EQ.1) THEN
            IF(K.EQ.NSEC.AND.I.GT.(INT(NCOL/2)+1)) GOTO 8
            IF(J.EQ.NROW.AND.I.GT.(INT(NCOL/2)+1)) GOTO 8
            ENDIF
            RADDSQ = (I-XC)**2 + YSQ + ZSQ
            IF(RADDSQ.GE.RADLOSQ .AND. RADDSQ.LE.RADHISQ) THEN
              B = MAP(I,J,K) 
            CALL MAG3D_INTRP (A, MAP2, NCOL1, NROW1, NSEC1, I, J,
     1        K, MAG, XC, YC, ZC)
            isOK = 0
            if(useAbs.eq.1) then
               if (abs(b-a) .lt. absThreshValue) isOK = 1
            else
               if((B-A) .gt. threshValue ) isOK = 1
            endif
            if(isOK .eq. 1) THEN
                BIN = INT((SQRT(RADDSQ) - RADLO) / RADSTEP) + 1
                IF(BIN.GT.NBINS) GOTO 8 
                  IIC = IIC + 1
C   DEBUGGING
              IF(IIC.GT.NFIT) THEN
                  PRINT *, 'IIC IS GREATER THAN NFIT!!'
                  CALL EXIT
              ENDIF
C   END DEBUGGING
                  CNTR1 (IIC) = B
                  CNTR2 (IIC) = A 
                SUMA(BIN)   = SUMA(BIN)   + A
                SUMB(BIN)   = SUMB(BIN)   + B
                SUMAB(BIN)  = SUMAB(BIN)  + A*B
                SUMASQ(BIN) = SUMASQ(BIN) + A*A
                SUMBSQ(BIN) = SUMBSQ(BIN) + B*B
                NTOT(BIN)   = NTOT(BIN)   + 1.0
            endif
            ENDIF
8          ENDDO
        ENDDO
      ENDDO
      GO TO 300
C
 300        TSUMA   = 0.0
        TSUMB   = 0.0
        TSUMAB  = 0.0
        TSUMASQ = 0.0
        TSUMBSQ = 0.0
        TNTOT   = 0.0
C
Ca       IF(RADTAG.EQ.0) PRINT *, 'RADIUS, CCOEF FOR RADLO->RADHI' 
      DO 303 IB = 1,NBINS 
       CCOEF(IB) = 0.
       RADIB = RADLO + (IB-1)*RADSTEP
       IF (NTOT(IB).EQ.0.) THEN 
Ca           IF(RADTAG.EQ.0) PRINT *, RADIB,CCOEF(IB) 
           GOTO 303
       ENDIF
       IF ((SUMASQ(IB) - (SUMA(IB)*SUMA(IB))/NTOT(IB)).LE.0. .OR.
     1 (SUMBSQ(IB) - (SUMB(IB)*SUMB(IB))/NTOT(IB)).LE.0.) THEN 
Ca           IF(RADTAG.EQ.0) PRINT *, RADIB,CCOEF(IB)
         ELSE 
         CCOEF(IB) = (SUMAB(IB) - (SUMA(IB)*SUMB(IB))/NTOT(IB)) /
     1           (SQRT( SUMASQ(IB) - (SUMA(IB)*SUMA(IB))/NTOT(IB))*
     2            SQRT(SUMBSQ(IB) - (SUMB(IB)*SUMB(IB))/NTOT(IB)))
        TSUMA   = TSUMA   + SUMA(IB)    
        TSUMB   = TSUMB   + SUMB(IB)
        TSUMAB  = TSUMAB  + SUMAB(IB)   
        TSUMASQ = TSUMASQ + SUMASQ(IB) 
        TSUMBSQ = TSUMBSQ + SUMBSQ(IB)  
        TNTOT   = TNTOT   + NTOT(IB)  
Ca          IF(RADTAG.EQ.0) PRINT *, RADIB,CCOEF(IB)
       ENDIF
 303      CONTINUE
      TCCOEF = (TSUMAB - (TSUMA*TSUMB)/TNTOT) /
     1           (SQRT(TSUMASQ - (TSUMA*TSUMA)/TNTOT)*
     2            SQRT(TSUMBSQ - (TSUMB*TSUMB)/TNTOT))
C

      deallocate(rdens_num, rdens_std)

      CALL CHI2FIT (CNTR2, CNTR1, IIC, MAP_STD, 1, AA, BB, UNA, UNB, 
     1 CHISQ, GDNS)
Ca      D0 = NUM_DIGIT_R4(MAG,4)
Ca      D1 = NUM_DIGIT_R4(AA,6)
Ca      D2 = NUM_DIGIT_R4(UNA,6)
Ca      D3 = NUM_DIGIT_R4(BB,6)
Ca      D4 = NUM_DIGIT_R4(UNB,6)
Ca      D5 = NUM_DIGIT_R4(TCCOEF,6)
Ca      D6 = NUM_DIGIT_R4(CHISQ,2)
Ca      PRINT *, ' FIT: [MAP1 = (SCALE*MAP2)',
Ca      1 ' + CONSTANT] ', ' ---------------------------------------- ' 
Ca      PRINT *, ' MAGNIFICATION (PIX_SIZE FACTOR): ', MAG
Ca      PRINT *, ' MAP1 = (MAP2 * ', BB, 
Ca      1 ') + (', AA, ')'
Ca      PRINT *, ' SCALE    = ', BB, ' +/-SD: ', 
Ca      1 UNB
Ca      PRINT *, ' CONSTANT = ', AA, ' +/-SD: ',
Ca      1 UNA
Ca      PRINT *, ' CCOEF, CHISQ: ', TCCOEF, CHISQ
C
C
C
 999      RETURN
      END
C
C*******************************************************************
C*  SUBROUTINE MAG3D_INTRP 
C*******************************************************************
      SUBROUTINE MAG3D_INTRP (A,MAP,NCOL,NROW,NSEC,I,J,K,MAG,
     1      XC, YC, ZC)
C
C      14-JAN-1992  BY HOLLAND CHENG 
C
      IMPLICIT NONE
C
      INTEGER*4 NCOL, NROW, NSEC, I, J, K, IX, IY, IZ 
      REAL*4    MAP(NCOL,NROW,NSEC)
      REAL*4    DX, DY, DZ, XC, YC, ZC, MAG, A, X, Y, Z
      REAL*4    D1, D2, D3, D4, D5, D6, D7, D8 
C  ------------------------------------------------------------
C
      A = 0.0
C
      X = MAG*(REAL(I)-XC) + XC
      Y = MAG*(REAL(J)-YC) + YC
      Z = MAG*(REAL(K)-ZC) + ZC
      IX = X
      IY = Y
      IZ = Z
Ca  I need this here for robem and doing 2D sections      
      if(iz .lt. 1) then
         iz = 1
      endif
      IF(IX.LT.1 .OR. IY.LT.1 .OR. IZ.LT.1 .OR.
     1  IX.GT.NCOL .OR. IY.GT.NROW .OR. IZ.GT.NSEC) RETURN 
C
      D2        = 0.0
      D3        = 0.0
      D4        = 0.0
      D5        = 0.0
      D6        = 0.0
      D7        = 0.0
      D8        = 0.0
      D1        = MAP(IX,IY,IZ)
      IF(IX.LT.NCOL) THEN 
         D2 = MAP(IX+1,IY,IZ)
         IF(IY.LT.NROW) THEN
            D3 = MAP(IX,IY+1,IZ)
            D4 = MAP(IX+1,IY+1,IZ)
            IF(IZ.LT.NSEC) THEN
               D5 = MAP(IX,IY,IZ+1)
               D6 = MAP(IX+1,IY,IZ+1)
               D7 = MAP(IX,IY+1,IZ+1)
               D8 = MAP(IX+1,IY+1,IZ+1)
            ENDIF
         ENDIF
      ENDIF
      DX        = X-IX
      DY        = Y-IY
      DZ        = Z-IZ
C
      A  =     (1.-DX)*(1.-DY)*(1.-DZ)*D1 + DX *(1.-DY)*(1.-DZ)*D2 
     1       + (1.-DX)*    DY *(1.-DZ)*D3 + DX *    DY *(1.-DZ)*D4
     2       + (1.-DX)*(1.-DY)*    DZ *D5 + DX *(1.-DY)*    DZ *D6
     3       + (1.-DX)*    DY *    DZ *D7 + DX *    DY *    DZ *D8
C
      RETURN
      END
C
      SUBROUTINE CHI2FIT (X, Y, NDATA, SIG, MWT, A, B, SIGA, SIGB,
     1 CHI2, Q)
C
      IMPLICIT NONE
C
      INTEGER  NDATA, I, MWT
      REAL*4   X(NDATA), Y(NDATA), SX, SY, ST2, WT, SS, SXOSS, T
      REAL*4   SIG, A, B, SIGA, SIGB, CHI2, Q, SIGDAT, GAMMQ
C  ----------------------------------------------------------------------------
      SX  = 0.0
      SY  = 0.0
      ST2 = 0.0
      B   = 0.0
      IF(MWT.NE.0) THEN
         SS = 0.0
         DO I = 1,NDATA
             WT = 1./(SIG**2)
             SS = SS+WT
             SX = SX+X(I)*WT
             SY = SY+Y(I)*WT
         ENDDO
       ELSE
           DO I = 1,NDATA
             SX = SX+X(I)
             SY = SY+Y(I)
         ENDDO
           SS = REAL(NDATA)
      ENDIF
C
      SXOSS = SX/SS
C
      IF(MWT.NE.0) THEN
         DO I = 1,NDATA
           T   = (X(I) - SXOSS)/SIG
           ST2 = ST2 + T*T
           B   = B + T*Y(I)/SIG
         ENDDO
       ELSE
         DO I = 1,NDATA
           T   = X(I) - SXOSS
           ST2 = ST2 + T*T
           B   = B + T*Y(I)
         ENDDO
      ENDIF
C
      B    = B/ST2
      A    = (SY-SX*B)/SS
      SIGA = SQRT((1.+SX*SX/(SS*ST2))/SS)
      SIGB = SQRT(1./ST2)
      CHI2 = 0.0
C
      IF(MWT.EQ.0) THEN
         DO I = 1,NDATA
           CHI2 = CHI2+(Y(I)-A-B*X(I))**2
         ENDDO
         Q      = 1.0
         SIGDAT = SQRT(CHI2/(NDATA-2))
         SIGA   = SIGA*SIGDAT
         SIGB   = SIGB*SIGDAT
       ELSE
         DO I=1,NDATA
           CHI2 = CHI2 + ((Y(I)-A-B*X(I))/SIG)**2
         ENDDO
c        Q=GAMMQ(0.5*(NDATA-2),0.5*CHI2)
         q = 1.
      ENDIF
C
      RETURN
      END



C
      SUBROUTINE INDEXX_DML(N,ARRIN,INDX)
      DIMENSION ARRIN(N),INDX(N)
      DO 11 J=1,N
        INDX(J)=J
11    CONTINUE
      IF(N.EQ.1)RETURN
      L=N/2+1
      IR=N
10    CONTINUE
        IF(L.GT.1)THEN
          L=L-1
          INDXT=INDX(L)
          Q=ARRIN(INDXT)
        ELSE
          INDXT=INDX(IR)
          Q=ARRIN(INDXT)
          INDX(IR)=INDX(1)
          IR=IR-1
          IF(IR.EQ.1)THEN
            INDX(1)=INDXT
            RETURN
          ENDIF
        ENDIF
        I=L
        J=L+L
20      IF(J.LE.IR)THEN
          IF(J.LT.IR)THEN
            IF(ARRIN(INDX(J)).LT.ARRIN(INDX(J+1)))J=J+1
          ENDIF
          IF(Q.LT.ARRIN(INDX(J)))THEN
            INDX(I)=INDX(J)
            I=J
            J=J+J
          ELSE
            J=IR+1
          ENDIF
        GO TO 20
        ENDIF
        INDX(I)=INDXT
      GO TO 10
      END
C
      SUBROUTINE MOMENT295 (DATA, N, AVE, ADEV, SDEV, VAR, SKEW, CURT)
C
      IMPLICIT NONE
      INTEGER  N,J
      REAL*4   DATA(N),S,AVE,ADEV,SDEV,VAR,SKEW,CURT,P
C
      IF(N.LE.1) THEN
         write(*,*) 'N must be at least 2'
         stop
      endif
      S=0.
      DO 11 J=1,N
        S=S+DATA(J)
11    CONTINUE
      AVE=S/N
      ADEV=0.
      VAR=0.
      SKEW=0.
      CURT=0.
      DO 12 J=1,N
        S=DATA(J)-AVE
        ADEV=ADEV+ABS(S)
        P=S*S
        VAR=VAR+P
        P=P*S
        SKEW=SKEW+P
        P=P*S
        CURT=CURT+P
12    CONTINUE
      ADEV=ADEV/N
      VAR=VAR/(N-1)
      SDEV=SQRT(VAR)
      IF(VAR.ne.0.)THEN
        SKEW=SKEW/(N*SDEV**3)
        CURT=CURT/(N*VAR**2)-3.  
      ELSE
cc        PAUSE 'no skew or kurtosis when zero variance'
        print *, 'no skew/kurtosis was calc when variance is to small'
        print *, '02-feb-1995[rhc]'
      ENDIF
      RETURN
      END
C  **********************************************************************
C  *                                                                    *
C  *                *****  SUBROUTINE CMP_MAP_1D  *****                 *
C  *                                                                    *
C  *    COMPUTE SPHERICALLY-AVERAGED DENSITY                              *
C  *                                                                    *
C  *    MAY 18, 2000 PURDUE UNIVERSITY ALPHASTATION 500, TSB,JTF        *
C  *                                                                    *
C  *    MODIFIED 18-MAY-2000 BY TSB/JTF (Bastardized version of MAP_1D) *
C  **********************************************************************
C
      SUBROUTINE CMP_MAP_1D (MAP, NCOL, NROW, NSEC, XC, YC, ZC,
     1            NRDENS, RSQ_LIM, RDENS, WRDENS, RDENS_NUM, RDENS_STD,
     2   QTR_SWITCH)
C
      IMPLICIT NONE
C
      include '../include/vax_minmax.inc'
C
      INTEGER*4    NCOL, NROW, NSEC, INUM, IR, IPUCK, QTR_SWITCH, NRDENS
      INTEGER*4    I, J, J1, J2, K, IXTAB, IYTAB, RDENS_NUM(NRDENS)
      REAL*4       XC, YC, ZC, RDENS(NRDENS), RDENS_MIN, RDENS_MAX
      REAL*4       WRDENS(NRDENS), SCALEFAC      
      REAL*4       MAP(NCOL,NROW,NSEC), X, Y, Z, YSQ, ZSQ
      REAL*4       TEST, RSQ_LIM, FRACTION, RDENS_STD(NRDENS)
      LOGICAL      OUTFILE, PLOT
      CHARACTER*64 OUT_FILENAME
C -----------------------------------------------------------------------------
C
      INUM    = NRDENS
C
      DO I = 1,INUM                      ! INITIALIZE 1D INTERPOLATION ARRAY
        RDENS_NUM(I) = 0
        RDENS(I)  = 0.0
        RDENS_STD(I) = 0.0
      ENDDO
C  **************************************************************
C  *    STEP THROUGH IMAGE AND REDUCE TO 1D CIRCULAR AVERAGE    *
C  **************************************************************
      DO K = 1,NSEC
        Z   = REAL(K) - ZC
        ZSQ = Z*Z
        DO J = 1,NROW
          Y   = REAL(J) - YC
          YSQ = Y*Y
          DO I = 1,NCOL
             IF(QTR_SWITCH.EQ.1) THEN
              IF(K.EQ.NSEC.AND.I.GT.(INT(NCOL/2)+1)) GOTO 5
              IF(J.EQ.NROW.AND.I.GT.(INT(NCOL/2)+1)) GOTO 5
             ENDIF
             X    = REAL(I) - XC
             TEST = X*X + YSQ + ZSQ
             IF(TEST.LE.RSQ_LIM) THEN
               IR             = NINT(SQRT(TEST)) + 1
             IF (IR.GT.NRDENS) GOTO 5 
               RDENS_NUM(IR)      = RDENS_NUM(IR) + 1
               RDENS(IR)       = RDENS(IR) + MAP(I,J,K)
             ENDIF
5          ENDDO
        ENDDO
      ENDDO
C  ********************************
C  *    NORMALIZE THE 1D ARRAY    *
C  ********************************
      RDENS_MIN = R4_MAX
      RDENS_MAX = R4_MIN
      DO I = 1,INUM
        IF(RDENS_NUM(I).GT.1) THEN
          RDENS(I)     = RDENS(I)/REAL(RDENS_NUM(I))            ! Get average
        ENDIF
        RDENS_MIN = MIN(RDENS_MIN,RDENS(I))
        RDENS_MAX = MAX(RDENS_MAX,RDENS(I))
      ENDDO

C  ******************************************************
C  *  NORMALIZE RDENS TO 1.0, REMOVE NEGATIVE DENSITY,  *
C  *  SAVE THIS WEIGHTED VALUE AS WRDENS                *
C  ******************************************************
      SCALEFAC = 1.0 / RDENS_MIN
      DO 10 I = 1,INUM
        WRDENS(I) = RDENS(I) * SCALEFAC 
        IF(RDENS(I).GT.0.0) THEN
          WRDENS(I) = 0.0
        ENDIF
10      ENDDO 

C  ***************************************************************
C  *  Compute standard deviation at each radius where RDENS_NUM > 1  *
C  ***************************************************************
      DO K = 1,NSEC
        Z   = REAL(K) - ZC
        ZSQ = Z*Z
        DO J = 1,NROW
          Y   = REAL(J) - YC
          YSQ = Y*Y
          DO I = 1,NCOL
             IF(QTR_SWITCH.EQ.1) THEN
              IF(K.EQ.NSEC.AND.I.GT.(INT(NCOL/2)+1)) GOTO 55
              IF(J.EQ.NROW.AND.I.GT.(INT(NCOL/2)+1)) GOTO 55
             ENDIF
             X    = REAL(I) - XC
             TEST = X*X + YSQ + ZSQ
             IF(TEST.LE.RSQ_LIM) THEN
               IR        = NINT(SQRT(TEST)) + 1
             IF(IR.GT.NRDENS) GOTO 55
               RDENS_STD(IR) = RDENS_STD(IR) + (RDENS(IR)-MAP(I,J,K))**2
             ENDIF
55          ENDDO
        ENDDO
      ENDDO
      DO I = 1,INUM
        IF (RDENS_NUM(I).GT.1) THEN
           RDENS_STD(I) = SQRT( RDENS_STD(I) / REAL(RDENS_NUM(I)-1) )
         ELSE
           RDENS_STD(I) = 0.0
        ENDIF
      ENDDO
C
      RETURN
      END
C
C****************************************************************************
C*  SUBROUTINE compute_scaling
C****************************************************************************
      SUBROUTINE COMPUTE_SCALING (MAP, MAP2, NCOL1, NROW1, NSEC1, 
     1 TBCCOEF, NCOL2, NROW2, NSEC2, NFIT, CNTR1, CNTR2, LPOL, 
     2 MAP_STD, XC,YC,ZC,RADLO,RADHI,RADSTEP,BMAG,
     3 ccoef,nbins, BB, AA, win_size, threshValue, useAbs)
C
      INTEGER*4 NCOL1, NROW1, NSEC1, NCOL2, NROW2, NSEC2
      INTEGER*4 I, J, K, L, NUM, NCOL, NROW, NSEC, OFFSET
      INTEGER*4 NMAG, II, IIX, IIY, IIZ, LIM, NBINS, BIN, IIC
      INTEGER*4 LPOL, RADTAG, NUM_DIGIT_R4, IB, NFIT, MWT
      INTEGER*4 D0, D1, D2, D3, D4, D5, D6
      INTEGER*4 WIN_SIZE, useAbs, isOK
      INTEGER*4 BUNCH, MAX_BUNCH
      PARAMETER (LIM=501)
      REAL*4    MAX_BUNCH_RADLO, MAX_BUNCH_RADHI
      REAL*4    MAX_BUNCH_PEAK, MAX_BUNCH_PEAKIB
      REAL*4    CCOEF(nbins), threshValue, absThreshValue
      REAL*4    RSQ_LIM, WCCOEF_MAX, S_WCCOEF(LIM)      
      REAL*4    CNTR1(NFIT), CNTR2(NFIT), MAG, MAP_STD
      REAL*4    AA, BB, UNA, UNB, CHISQ, GDNS, TCCOEF, TBCCOEF, BMAG
      REAL*4    AVE,ADEV,SDEV,VAR,SKEW,CURT,MAGCCOEF(LIM)
      REAL*4    RADLO,RADHI,RADDSQ,XC,YC,ZC,RADLOSQ,RADHISQ,YSQ,ZSQ
      REAL*4    MAP(NCOL1,NROW1,NSEC1),A,B,RADSTEP
      REAL*4    MAP2(NCOL2,NROW2,NSEC2), S, P
      REAL*4    TNTOT, NTOT(LIM), RADIB, TMP1, TMP2
      REAL*8    TSUMA, TSUMB, TSUMASQ, TSUMBSQ, TSUMAB, SUMAB(LIM)
      REAL*8    SUMA(LIM), SUMB(LIM), SUMASQ(LIM), SUMBSQ(LIM)
C  -------------------------------------------------------------------
C
Ca      RWA - for now let's just init these - we'll pas them in later???
      absThreshValue = abs(threshValue)
      iix = 1
      iiy = 1
      iiz = 1
      MAG     = BMAG 
      RADTAG  = 0 
      NCOL    = MAX (NCOL1, NCOL2)
      NROW    = NCOL
      NSEC    = NCOL
      WCCOEF_MAX = 0.0
C
      NBINS = 1 + NINT((RADHI-RADLO)/RADSTEP)
      IF(NBINS.LE.0) NBINS = 1
      RADLOSQ = RADLO*RADLO
      RADHISQ = (RADHI+RADSTEP)**2 
      DO IB = 1,NBINS
        SUMA(IB)    = 0.0
        SUMB(IB)    = 0.0
        SUMAB(IB)   = 0.0
        SUMASQ(IB)  = 0.0
        SUMBSQ(IB)  = 0.0
        NTOT(IB)    = 0.0
      ENDDO
C
      IIC = 0 
C      IF(NCOL1.NE.NCOL2) GO TO 100
C  *******************************************
C  *    HERE IF NCOL1=NCOL2 AND NROW1=NROW2  *
C  *******************************************
      DO K = iiz,NSEC
        ZSQ = (K-ZC)**2
        DO J = iiy,NROW
          YSQ = (J-YC)**2
          DO I = iix,NCOL
            RADDSQ = (I-XC)**2 + YSQ + ZSQ
            IF(RADDSQ.GE.RADLOSQ .AND. RADDSQ.LE.RADHISQ) then
              B = MAP(I,J,K) 
            CALL MAG3D_INTRP (A, MAP2, NCOL, NROW, NSEC, I, J,
     1        K, MAG, XC, YC, ZC)
            isOK = 0
            if(useAbs.eq.1) then
               if (abs(b-a) .lt. absThreshValue) isOK = 1
            else
               if((B-A) .gt. threshValue ) isOK = 1
            endif
            if(isOK .eq. 1) THEN
                BIN = INT((SQRT(RADDSQ) - RADLO) / RADSTEP) + 1
                IF(BIN.GT.NBINS) GOTO 8 
                  IIC = IIC + 1
                  CNTR1 (IIC) = B
                  CNTR2 (IIC) = A 
                SUMA(BIN)   = SUMA(BIN)   + A
                SUMB(BIN)   = SUMB(BIN)   + B
                SUMAB(BIN)  = SUMAB(BIN)  + A*B
                SUMASQ(BIN) = SUMASQ(BIN) + A*A
                SUMBSQ(BIN) = SUMBSQ(BIN) + B*B
                NTOT(BIN)   = NTOT(BIN)   + 1.0
            endif  
            ENDIF
8          ENDDO
        ENDDO
      ENDDO
      GO TO 300
C
 300        TSUMA   = 0.0
        TSUMB   = 0.0
        TSUMAB  = 0.0
        TSUMASQ = 0.0
        TSUMBSQ = 0.0
        TNTOT   = 0.0
C
Ca       IF(RADTAG.EQ.0) WRITE(*,*) 'RADIUS, CCOEF FOR RADLO->RADHI' 
      DO 303 IB = 1,NBINS 
       CCOEF(IB) = 0.
       RADIB = RADLO + (IB-1)*RADSTEP
       IF (NTOT(IB).EQ.0.) THEN 
Ca           IF(RADTAG.EQ.0) WRITE(*,'(2F10.4,e12.4)') RADIB,CCOEF(IB) 
           GOTO 303
       ENDIF
       IF ((SUMASQ(IB) - (SUMA(IB)*SUMA(IB))/NTOT(IB)).LE.0. .OR.
     1 (SUMBSQ(IB) - (SUMB(IB)*SUMB(IB))/NTOT(IB)).LE.0.) THEN 
Ca           IF(RADTAG.EQ.0) WRITE(*,'(2F10.4,e12.4)') RADIB,CCOEF(IB)
         ELSE 
         CCOEF(IB) = (SUMAB(IB) - (SUMA(IB)*SUMB(IB))/NTOT(IB)) /
     1           (SQRT( SUMASQ(IB) - (SUMA(IB)*SUMA(IB))/NTOT(IB))*
     2            SQRT(SUMBSQ(IB) - (SUMB(IB)*SUMB(IB))/NTOT(IB)))
        TSUMA   = TSUMA   + SUMA(IB)    
        TSUMB   = TSUMB   + SUMB(IB)
        TSUMAB  = TSUMAB  + SUMAB(IB)   
        TSUMASQ = TSUMASQ + SUMASQ(IB) 
        TSUMBSQ = TSUMBSQ + SUMBSQ(IB)  
        TNTOT   = TNTOT   + NTOT(IB)  
Ca          IF(RADTAG.EQ.0) WRITE(*,'(2F10.4,e12.4)') RADIB,CCOEF(IB)
       ENDIF
 303      CONTINUE
      if(tntot .gt. 0.0) then
      TCCOEF = (TSUMAB - (TSUMA*TSUMB)/TNTOT) /
     1           (SQRT(TSUMASQ - (TSUMA*TSUMA)/TNTOT)*
     2            SQRT(TSUMBSQ - (TSUMB*TSUMB)/TNTOT))
C
      CALL CHI2FIT (CNTR2, CNTR1, IIC, MAP_STD, 1, AA, BB, UNA, UNB, 
     1 CHISQ, GDNS)
C
      D0 = NUM_DIGIT_R4(MAG,4)
      D1 = NUM_DIGIT_R4(AA,6)
      D2 = NUM_DIGIT_R4(UNA,6)
      D3 = NUM_DIGIT_R4(BB,6)
      D4 = NUM_DIGIT_R4(UNB,6)
      D5 = NUM_DIGIT_R4(TCCOEF,6)
      D6 = NUM_DIGIT_R4(CHISQ,2)
      endif
Ca      WRITE(*,'(/,2A,/,A,/,A,F<D0>.4)') ' FIT: [MAP1 = (SCALE*MAP2)',
Ca      1 ' + CONSTANT] ', ' ---------------------------------------- ', 
Ca      2 ' MAGNIFICATION (PIX_SIZE FACTOR): ', MAG
Ca      WRITE(*,'(A,F<D3>.6,A,F<D1>.6,A)') ' MAP1 = (MAP2 * ', BB, 
Ca      1 ') + (', AA, ')'
Ca      WRITE(*,'(A,F<D3>.6,A,F<D4>.6)') ' SCALE    = ', BB, ' +/-SD: ', 
Ca      1 UNB
Ca      WRITE(*,'(A,F<D1>.6,A,F<D2>.6)') ' CONSTANT = ', AA, ' +/-SD: ',
Ca      1 UNA
c      IF(MWT.EQ.1) PRINT *,' (BASED ON MAP_STD:', MAP_STD,')'
Ca      WRITE(*,'(A,F<D5>.6,2X,F<D6>.2,/)') ' CCOEF, CHISQ: ', TCCOEF, CHISQ
C
Ca      WRITE(3,'(/,2A,/,A,/,A,F<D0>.4)') ' FIT: [(MAP1) = SCALE*(MAP2)',
Ca      1 ' + CONSTANT] ', ' ---------------------------------------- ', 
Ca      2 ' MAGNIFICATION (PIX_SIZE FACTOR): ', MAG
Ca      WRITE(3,'(A,F<D3>.6,A,F<D1>.6)') ' MAP1 = MAP2 * ', BB, ' + ', AA
Ca      WRITE(3,'(A,F<D3>.6,A,F<D4>.6)') ' SCALE    = ', BB, ' +/-SD: ', 
Ca      1 UNB
Ca      WRITE(3,'(A,F<D1>.6,A,F<D2>.6)') ' CONSTANT = ', AA, ' +/-SD: ', 
Ca      1 UNA
c      IF(MWT.EQ.1) PRINT *,' (BASED ON MAP_STD:', MAP_STD,')'
Ca      WRITE(3,'(A,F<D5>.6,2X,F<D6>.2,/)') ' CCOEF, CHISQ: ', TCCOEF, CHISQ
C
C
 999      RETURN
      END
